"""Convert a string to a `pathlib.Path` object."""

# pylint: disable=too-many-locals
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
# pylint: disable=too-many-arguments

import fnmatch
import os
from pathlib import Path
from typing import Any

from click_extended.core.nodes.child_node import ChildNode
from click_extended.core.other.context import Context
from click_extended.types import Decorator


class ToPath(ChildNode):
    """Convert a string to a `pathlib.Path` object."""

    def handle_str(
        self, value: str, context: Context, *args: Any, **kwargs: Any
    ) -> Path:
        parent = context.get_current_parent_as_parent()
        name = parent.name
        exists = kwargs["exists"]
        parents = kwargs["parents"]
        extensions = kwargs["extensions"]
        include_pattern = kwargs["include_pattern"]
        exclude_pattern = kwargs["exclude_pattern"]
        allow_file = kwargs["allow_file"]
        allow_directory = kwargs["allow_directory"]
        allow_empty_directory = kwargs["allow_empty_directory"]
        allow_empty_file = kwargs["allow_empty_file"]
        allow_symlink = kwargs["allow_symlink"]
        follow_symlink = kwargs["follow_symlink"]
        resolve = kwargs["resolve"]
        is_readable = kwargs["is_readable"]
        is_writable = kwargs["is_writable"]
        is_executable = kwargs["is_executable"]

        path = Path(value) if isinstance(value, str) else value
        path = path.expanduser()

        # Symlinks
        if not allow_symlink and path.is_symlink():
            raise OSError(
                f"Path '{path}' is a symlink, but symlinks are not allowed "
                f"for '{name}'"
            )

        if resolve:
            path = path.resolve() if follow_symlink else path.absolute()
        elif follow_symlink and path.is_symlink():
            path = path.resolve()
        else:
            path = path.absolute()

        # Existence
        if exists and not path.exists():
            raise FileNotFoundError(f"Path '{path}' does not exist.")

        # Parents
        if parents and not path.parent.exists():
            path.parent.mkdir(parents=True, exist_ok=True)

        if path.exists():
            is_file = path.is_file()
            is_dir = path.is_dir()

            # File disallowed
            if is_file and not allow_file:
                raise IsADirectoryError(
                    f"Path '{path}' is a file, but files are not allowed "
                    f"for '{name}'"
                )

            # Directory disallowed
            if is_dir and not allow_directory:
                raise NotADirectoryError(
                    f"Path '{path}' is a directory, but directories are not "
                    f"allowed for '{name}'"
                )

            # Check empty directory
            if is_dir and not allow_empty_directory:
                if not any(path.iterdir()):
                    raise ValueError(
                        f"Directory '{path}' is empty, but empty directories "
                        f"are not allowed for '{name}'"
                    )

            # Check empty file
            if is_file and not allow_empty_file:
                if path.stat().st_size == 0:
                    raise ValueError(
                        f"File '{path}' is empty, but empty files are not "
                        f"allowed for '{name}'"
                    )

            # Check permissions
            if is_readable and not os.access(path, os.R_OK):
                raise PermissionError(
                    f"Path '{path}' is not readable for '{name}'"
                )

            if is_writable and not os.access(path, os.W_OK):
                raise PermissionError(
                    f"Path '{path}' is not writable for '{name}'"
                )

            if is_executable and not os.access(path, os.X_OK):
                raise PermissionError(
                    f"Path '{path}' is not executable for '{name}'"
                )

        # Check extensions
        if extensions and (not path.exists() or path.is_file()):
            extensions = [
                ext if ext.startswith(".") else f".{ext}" for ext in extensions
            ]

            if not any(path.name.endswith(ext) for ext in extensions):
                raise ValueError(
                    f"Path '{path}' does not have an allowed extension. "
                    f"Allowed extensions: {', '.join(extensions)} "
                    f"for '{name}'"
                )

        # Patterns
        if include_pattern:
            if not fnmatch.fnmatch(path.name, include_pattern):
                raise ValueError(
                    f"Path '{path}' does not match include pattern "
                    f"'{include_pattern}' for '{name}'"
                )
        elif exclude_pattern and fnmatch.fnmatch(path.name, exclude_pattern):
            raise ValueError(
                f"Path '{path}' matches exclude pattern "
                f"'{exclude_pattern}' for '{name}'"
            )

        return path


def to_path(
    *,
    exists: bool = True,
    parents: bool = False,
    resolve: bool = True,
    extensions: list[str] | None = None,
    include_pattern: str | None = None,
    exclude_pattern: str | None = None,
    allow_file: bool = True,
    allow_directory: bool = True,
    allow_empty_directory: bool = True,
    allow_empty_file: bool = True,
    allow_symlink: bool = False,
    follow_symlink: bool = True,
    is_readable: bool = False,
    is_writable: bool = False,
    is_executable: bool = False,
) -> Decorator:
    """
    Convert, validate, and process a string to a `pathlib.Path` object.

    Type: `ChildNode`

    Supports: `str`

    Args:
        exists (bool, optional):
            Whether the path needs to exist.
            Defaults to `True`.

        parents (bool, optional):
            Create parent directories if they don't exist.
            Defaults to `False`.

        resolve (bool, optional):
            Convert the path to an absolute path. When `True`, resolves
            `.` and `..` components. When False, only makes the path
            absolute without resolution.
            Defaults to `True`.

        extensions (list[str], optional):
            A list of extensions to require the path to end with.
            By default, all extensions are allowed.

        include_pattern (str, optional):
            A whitelist pattern. Uses shell-style glob patterns.
            Defaults to `None` (All file names allowed).

            When both `include_pattern` and `exclude_pattern` are provided,
            `include_pattern` takes precedence (files matching include
            are always accepted).

            ```python
            include_pattern="*.py"      # Only Python files
            include_pattern="config_*"  # Files starting with "config_"
            ```

        exclude_pattern (str, optional):
            A blacklist pattern. Uses shell-style glob patterns.
            Defaults to `None` (No file names excluded).

            If provided without `include_pattern`, file names matching this
            pattern will be rejected, and all other file names will be
            allowed.

            When both `include_pattern` and `exclude_pattern` are provided,
            `include_pattern` takes precedence (files matching include
            are always accepted).

        allow_file (bool, optional):
            Whether to allow the path to be a file or not.
            Defaults to `True`.

        allow_directory (bool, optional):
            Whether to allow the path to be a directory or not.
            Defaults to `True`.

        allow_empty_directory (bool, optional):
            If the path points to a directory, whether to allow
            the directory to be empty. Only checked when the path
            is a directory. Defaults to `True`.

        allow_empty_file (bool, optional):
            If the path points to a file, whether to allow the
            file to be empty. Only checked when the path is a file.
            Defaults to `True`.

        allow_symlink (bool, optional):
            Whether to allow the path to be a symlink or not.
            Defaults to `False`.

        follow_symlink (bool, optional):
            Whether to follow symlinks when resolving the path.
            Only applies when resolve=True.
            Defaults to `True`.

        is_readable (bool, optional):
            Whether the file has `read` permissions.
            Defaults to `False`.

        is_writable (bool, optional):
            Whether the file has `write` permissions.
            Default to `False`.

        is_executable (bool, optional):
            A unix-only feature that checks if the file has
            `execute` permissions. Defaults to `False`.

    Returns:
        Decorator:
            The decorated function.
    """
    return ToPath.as_decorator(
        exists=exists,
        parents=parents,
        resolve=resolve,
        extensions=extensions,
        include_pattern=include_pattern,
        exclude_pattern=exclude_pattern,
        allow_file=allow_file,
        allow_directory=allow_directory,
        allow_empty_directory=allow_empty_directory,
        allow_empty_file=allow_empty_file,
        allow_symlink=allow_symlink,
        follow_symlink=follow_symlink,
        is_readable=is_readable,
        is_writable=is_writable,
        is_executable=is_executable,
    )



"""Child decorator to convert a string to a datetime."""

from datetime import datetime
from typing import Any
from zoneinfo import ZoneInfo

from click_extended.core.nodes.child_node import ChildNode
from click_extended.core.other.context import Context
from click_extended.types import Decorator
from click_extended.utils import humanize_iterable
from click_extended.utils.time import normalize_datetime_format


class ToDatetime(ChildNode):
    """Child decorator to convert a string to a datetime."""

    def handle_str(
        self,
        value: str,
        context: Context,
        *args: Any,
        **kwargs: Any,
    ) -> datetime:
        formats = kwargs["formats"] or (
            "%Y-%m-%d",
            "%H:%M:%S",
            "%Y-%m-%d %H:%M:%S",
        )
        tz = kwargs.get("tz")

        for fmt in formats:
            try:
                normalized_fmt = normalize_datetime_format(fmt)
                dt = datetime.strptime(value, normalized_fmt)
                if tz:
                    dt = dt.replace(tzinfo=ZoneInfo(tz))
                return dt
            except ValueError:
                continue

        fmt = "either of the formats" if len(formats) != 1 else "in the format"
        raise ValueError(
            f"Invalid datetime '{value}', must be in "
            f"{fmt} {humanize_iterable(formats, sep='or')}"
        )


def to_datetime(
    *formats: str,
    tz: str | None = None,
) -> Decorator:
    """
    Convert a string to a datetime by trying multiple formats.

    Type: `ChildNode`

    Supports: `str`

    Args:
        *formats (str):
            One or more datetime format strings to try. Supports both Python
            strptime format (e.g., "%Y-%m-%d", "%d/%m/%Y") and simplified format
            (e.g., "YYYY-MM-DD", "DD/MM/YYYY"). The decorator will attempt each
            format in order until one succeeds. Defaults to `"%Y-%m-%d"`,
            `"%H:%M:%S"` and `"%Y-%m-%d %H:%M:%S"`,

        tz (str | None, optional):
            Timezone name (e.g., "UTC", "America/New_York", "Europe/Stockholm")
            to apply to the parsed datetime. Uses zoneinfo.ZoneInfo for timezone
            handling. Defaults to `None` (naive datetime).

    Returns:
        Decorator:
            The decorated function.

    Example:
        @to_datetime("YYYY-MM-DD", "DD/MM/YYYY", tz="America/New_York")
        # Or using Python strptime format:
        @to_datetime("%Y-%m-%d", "%d/%m/%Y", tz="America/New_York")
        def process_date(date: datetime):
            print(date)
    """
    return ToDatetime.as_decorator(formats=formats, tz=tz)



"""Hook event payload definitions."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import click

from click_extended.hooks.hook_phase import HookPhase

if TYPE_CHECKING:
    from click_extended.core.nodes._root_node import RootNode
    from click_extended.core.other.context import Context


@dataclass(frozen=True)
class HookEvent:
    """
    Data passed to hook handlers.

    :param phase: The lifecycle phase that triggered the hook.
    :param click_context: Active Click context.
    :param root: Root node of the CLI, if scoped.
    :param context: click-extended context, if initialized.
    :param exception: Exception that triggered the hook, if any.
    """

    phase: HookPhase
    click_context: click.Context
    root: "RootNode | None"
    context: "Context | None"
    exception: BaseException | None = None



"""Convert between different power units."""

from decimal import Decimal, getcontext
from typing import Any, Literal

from click_extended.core.nodes.child_node import ChildNode
from click_extended.core.other.context import Context
from click_extended.types import Decorator

getcontext().prec = 35

UNITS = {
    "W": Decimal("1"),
    "kW": Decimal("1e3"),
    "MW": Decimal("1e6"),
    "GW": Decimal("1e9"),
    "TW": Decimal("1e12"),
    "hp": Decimal("745.69987158227022"),
    "hpM": Decimal("735.49875"),
    "ftlbs": Decimal("1.3558179483314004"),
    "Btuh": Decimal("0.2930710701722222"),
    "Btus": Decimal("1055.05585262"),
    "tonref": Decimal("3516.852842066666"),
}


class ConvertPower(ChildNode):
    """Convert between different power units."""

    def handle_numeric(
        self,
        value: int | float,
        context: Context,
        *args: Any,
        **kwargs: Any,
    ) -> float:
        from_unit = kwargs["from_unit"]
        to_unit = kwargs["to_unit"]
        val = Decimal(str(value))

        if from_unit == "dBW":
            watts = Decimal(10) ** (val / 10)
        elif from_unit == "dBm":
            watts = Decimal(10) ** ((val - 30) / 10)
        elif from_unit in UNITS:
            watts = val * UNITS[from_unit]
        else:
            raise ValueError(f"Unknown unit '{from_unit}'")

        if to_unit == "dBW":
            if watts <= 0:
                raise ValueError(
                    "Power must be positive for logarithmic conversion"
                )
            result = 10 * watts.log10()
        elif to_unit == "dBm":
            if watts <= 0:
                raise ValueError(
                    "Power must be positive for logarithmic conversion"
                )
            result = 10 * (watts * 1000).log10()
        elif to_unit in UNITS:
            result = watts / UNITS[to_unit]
        else:
            raise ValueError(f"Unknown unit '{to_unit}'")

        return float(result)


def convert_power(
    from_unit: Literal[
        "W",
        "kW",
        "MW",
        "GW",
        "TW",
        "hp",
        "hpM",
        "dBW",
        "dBm",
        "Btuh",
        "Btus",
        "ftlbs",
        "tonref",
    ],
    to_unit: Literal[
        "W",
        "kW",
        "MW",
        "GW",
        "TW",
        "hp",
        "hpM",
        "dBW",
        "dBm",
        "Btuh",
        "Btus",
        "ftlbs",
        "tonref",
    ],
) -> Decorator:
    """
    Convert between different power units.

    Type: `ChildNode`

    Supports: `int`, `float`

    Units:
        - **W**: Watt
        - **kW**: Kilowatt
        - **MW**: Megawatt
        - **GW**: Gigawatt
        - **TW**: Terawatt
        - **hp**: Horsepower
        - **hpM**: Metric horsepower
        - **dBW**: Decibel-watt
        - **dBm**: Decibel-milliwatt
        - **Btuh**: BTU per hour
        - **Btus**: BTU per second
        - **ftlbs**: Foot-pounds per second
        - **tonref**: Ton of refrigeration (cooling capacity)

    Returns:
        Decorator:
            The decorated function.
    """
    return ConvertPower.as_decorator(from_unit=from_unit, to_unit=to_unit)



"""`ParentNode` to inject the context into the function."""

# pylint: disable=redefined-builtin
# pylint: disable=redefined-outer-name

from typing import Any

from click_extended.core.nodes.parent_node import ParentNode
from click_extended.core.other.context import Context as Ctx
from click_extended.types import Decorator


class Context(ParentNode):
    """`ParentNode` to inject the context into the function."""

    def load(self, context: Ctx, *args: Any, **kwargs: Any) -> Ctx:
        return context


def context(
    name: str = "ctx",
    param: str | None = None,
    help: str | None = None,
    tags: str | list[str] | None = None,
    **kwargs: Any,
) -> Decorator:
    """
    A `ParentNode` to inject the context into the function.

    Type: `ParentNode`

    Args:
        name (str, optional):
            Internal node name (must be snake_case). Defaults
            to "ctx".
        param (str, optional):
            The parameter name to inject into the function.
            If not provided, uses `name` (or derived name).
        help (str, optional):
            Help text for this parameter.
        tags (str | list[str], optional):
            Tag(s) to associate with this parameter for grouping.
        **kwargs (Any):
            Additional keyword arguments.

    Returns:
        Callable:
            A decorator function that registers the context parent node.
    """
    return Context.as_decorator(
        name=name,
        param=param,
        help=help,
        tags=tags,
        **kwargs,
    )



"""Child decorator to remove a suffix from a string."""

from typing import Any

from click_extended.core.nodes.child_node import ChildNode
from click_extended.core.other.context import Context
from click_extended.types import Decorator


class RemoveSuffix(ChildNode):
    """Child decorator to remove a suffix from a string."""

    def handle_str(
        self, value: str, context: Context, *args: Any, **kwargs: Any
    ) -> str:
        suffix: str = kwargs["suffix"]

        if value.endswith(suffix):
            return value[: -len(suffix)]

        return value


def remove_suffix(suffix: str) -> Decorator:
    """
    Remove a suffix from a string if it exists.

    Type: `ChildNode`

    Supports: `str`

    Args:
        suffix (str):
            The suffix to remove from the end of the string.

    Returns:
        Decorator:
            The decorator function.

    Examples:
        ```python
        @command()
        @option("filename", default="document.txt")
        @remove_suffix(".txt")
        def cmd(filename: str) -> None:
            click.echo(f"Name: {filename}")  # Output: Name: document
        ```

        ```python
        @command()
        @option("url", default="example.com/")
        @remove_suffix("/")
        def cmd(url: str) -> None:
            click.echo(f"URL: {url}")  # Output: URL: example.com
        ```
    """
    return RemoveSuffix.as_decorator(suffix=suffix)



"""`ParentNode` that loads a value from an environment variable."""

# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=redefined-builtin

import os
from typing import Any, Callable, ParamSpec, TypeVar, cast

from dotenv import load_dotenv

from click_extended.core.nodes.parent_node import ParentNode
from click_extended.core.other.context import Context
from click_extended.utils.casing import Casing

load_dotenv()

P = ParamSpec("P")
T = TypeVar("T")


class Env(ParentNode):
    """`ParentNode` that loads a value from an environment variable."""

    def load(self, context: Context, *args: Any, **kwargs: Any) -> Any:
        """
        Load and return the environment variable value.

        Args:
            context (Context):
                The current context instance.
            *args (Any):
                Optional positional arguments.
            **kwargs (Any):
                Keyword arguments from the decorator, including:
                - env_name (str): The environment variable name to read.

        Returns:
            Any:
                The value of the environment variable, or the default value
                if not required and not set.

        Raises:
            ValueError:
                If the environment variable is required but not set.
        """
        env_name = kwargs.get("env_name")
        if env_name is None:
            raise ValueError("env_name must be provided")

        value = os.getenv(env_name)

        if value is None:
            if self.required:
                raise ValueError(
                    f"Required environment variable '{env_name}' " "is not set."
                )
            value = self.default

        return value

    def get_display_name(self) -> str:
        """
        Get a formatted display name for error messages.

        Returns:
            str:
                The environment variable name in SCREAMING_SNAKE_CASE.
        """
        return Casing.to_screaming_snake_case(self.name)

    def check_required(self) -> str | None:
        """
        Check if required environment variable is set.

        Returns:
            str | None:
                The name of the missing environment variable if required
                and not set, otherwise None.
        """
        env_name = self.decorator_kwargs.get("env_name")
        if env_name and self.required and os.getenv(env_name) is None:
            return cast(str, env_name)
        return None


def env(
    env_name: str,
    name: str | None = None,
    param: str | None = None,
    help: str | None = None,
    required: bool = False,
    default: Any = None,
    tags: str | list[str] | None = None,
    **kwargs: Any,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """
    A `ParentNode` decorator to inject an environment variable value
    into a command.

    Type: `ParentNode`

    Args:
        env_name (str):
            The name of the environment variable to read (e.g., "API_KEY").
            Can be in any format, typically SCREAMING_SNAKE_CASE.
        name (str, optional):
            Internal node name (must be snake_case). If not provided,
            uses env_name converted to snake_case.
        param (str, optional):
            The parameter name to inject into the function.
            If not provided, uses name (or derived name).
        help (str, optional):
            Help text for this parameter.
        required (bool):
            Whether this parameter is required. Defaults to `False`.
            If `True`, the environment variable must be set, even if
            a default value is provided. The default is ignored when
            `required=True`.
        default (Any):
            Default value if environment variable is not set and
            `required=False`. Defaults to `None`.
        tags (str | list[str], optional):
            Tag(s) to associate with this parameter for grouping.
        **kwargs (Any):
            Additional keyword arguments.

    Returns:
        Callable:
            A decorator function that registers the env parent node.

    Examples:
        >>> @env("API_KEY")
        ... def my_func(api_key):
        ...     print(api_key)

        >>> @env("DATABASE_URL", param="db", required=True)
        ... def my_func(db):
        ...     print(db)

        >>> @env("MY_ENV", param="api_key")
        ... def my_func(api_key):
        ...     print(api_key)
    """
    node_name = name if name is not None else Casing.to_snake_case(env_name)
    return Env.as_decorator(
        name=node_name,
        env_name=env_name,
        param=param,
        help=help,
        required=required,
        default=default,
        tags=tags,
        **kwargs,
    )



"""Check if a value is a valid hostname."""

import re
from typing import Any

from click_extended.core.nodes.child_node import ChildNode
from click_extended.core.other.context import Context
from click_extended.types import Decorator

HOSTNAME_PATTERN = re.compile(r"(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)


class IsHostname(ChildNode):
    """Check if a value is a valid hostname."""

    def handle_str(
        self,
        value: str,
        context: Context,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        if len(value) > 255:
            raise ValueError(f"Value '{value}' is too long to be a hostname.")

        if value[-1] == ".":
            value = value[:-1]

        if not all(HOSTNAME_PATTERN.match(x) for x in value.split(".")):
            raise ValueError(f"Value '{value}' is not a valid hostname.")

        return value


def is_hostname() -> Decorator:
    """
    Check if a value is a valid hostname.

    Type: `ChildNode`

    Supports: `str`

    Returns:
        Decorator:
            The decorated function.
    """
    return IsHostname.as_decorator()



"""Child node to add a prefix to a string."""

from typing import Any

from click_extended.core.nodes.child_node import ChildNode
from click_extended.core.other.context import Context
from click_extended.types import Decorator


class AddPrefix(ChildNode):
    """Child node to add a prefix to a string."""

    def handle_str(
        self, value: str, context: Context, *args: Any, **kwargs: Any
    ) -> Any:
        prefix = kwargs["prefix"]
        skip = kwargs["skip"]
        case_sensitive = kwargs["case_sensitive"]

        if skip:
            if case_sensitive:
                if value.startswith(prefix):
                    return value
            else:
                if value.lower().startswith(prefix.lower()):
                    return value

        return prefix + value


def add_prefix(
    prefix: str,
    skip: bool = True,
    case_sensitive: bool = False,
) -> Decorator:
    """
    Add a prefix to a string.

    Type: `ChildNode`

    Supports: `str`

    Args:
        prefix (str):
            The prefix to add.
        skip (bool, optional):
            Skip adding the prefix if it already exists. Defaults to `True`.
        case_sensitive (bool)
            Check for exact case matching when `skip=True`. Defaults to `False`.
    Returns:
        Decorator:
            The decorated function.
    """
    return AddPrefix.as_decorator(
        prefix=prefix,
        skip=skip,
        case_sensitive=case_sensitive,
    )
